#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import sysconfig
import setuptools
import build_pdfium

from _setup_base import (
    PlatformDirs,
    BDistBase,
    SetupKws,
    build,
)


class bdist (BDistBase):
    def finalize_options(self):
        BDistBase.finalize_options(self)
        self.plat_name = sysconfig.get_platform()


def lib_setup(libname):
    setuptools.setup(
        cmdclass = {'bdist_wheel': bdist},
        package_data = {'': [libname]},
        **SetupKws,
    )


def get_binary():
    libpath = build_pdfium.find_lib(directory=PlatformDirs.SourceBuild)
    return os.path.basename(libpath)


def main():
    libname = get_binary()
    build(lambda: lib_setup(libname), PlatformDirs.SourceBuild)


if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import io
import sys
import warnings
from os.path import abspath
from typing import (
    Union,
    Optional,
    BinaryIO,
)

from pypdfium2 import _pypdfium as pdfium
from pypdfium2._helpers.nativeopener import *
from pypdfium2._helpers.error_handler import *


class PdfContext:
    """
    Context manager to open and automatically close again a PDFium document.
    It internally uses :func:`.open_pdf_auto` and :func:`.close_pdf`.
    This is the safest and most convenient way to open a document using PyPDFium2.
    
    Parameters:
        input_obj:
            File path to a PDF document, bytes, or a byte buffer.
        password:
            A password to unlock the document, if encrypted.
    """
    
    def __init__(
            self,
            input_obj,
            password = None,
        ):
        self.input_obj = input_obj
        self.password = password
        self.ld_data = None
    
    def __enter__(self):
        self.pdf, self.ld_data = open_pdf_auto(self.input_obj, self.password)
        return self.pdf
    
    def __exit__(self, exc_type, exc_value, exc_traceback):
        close_pdf(self.pdf, self.ld_data)


def _str_isascii(string):
    try:
        tmp = string.encode('ascii')
    except UnicodeEncodeError:
        return False
    else:
        return True


def open_pdf_auto(
        input_obj: Union[str, bytes, BinaryIO],
        password: Union[str, bytes] = None,
    ) -> Tuple[pdfium.FPDF_DOCUMENT, Optional[LoaderData]]:
    """    
    Open a document from a file path or in-memory data. If you are not able to use the context
    manager :class:`.PdfContext`, this is the recommended function to use for document opening.
    
    If the input is a regular file path, ``FPDF_LoadDocument()`` will be used.
    If on Windows, file paths that contain non-ascii characters will be loaded using
    :func:`.open_pdf_native`.
    If the input is bytes or a byte buffer, :func:`.open_pdf_buffer` will be used.
    
    Parameters:
        input_obj:
            File path to a PDF document, bytes, or a byte buffer.
        password:
            A password to unlock the document, if encrypted.
    
    Returns:
        The handle to a PDFium document, and a :class:`.LoaderData` object to store associated
        file access data.
    
    Warning:
        Callers **MUST** ensure that the :class:`.LoaderData` object remain available for as
        long as they work with the PDF. That means it has to be accessed again when done with
        processing, to prevent Python from automatically deleting the object. This can be
        achieved by passing it as second parameter to :func:`.close_pdf`, which is also required
        to close the opened file buffer.
        If attempting to access the ``FPDF_DOCUMENT`` handle after the loader data has been deleted,
        a segmentation fault would occur.
    """
    
    if isinstance(input_obj, bytes):
        input_obj = io.BytesIO(input_obj)
    
    
    ld_data = None
    
    if isinstance(input_obj, str):
        if sys.platform.startswith('win32') and not _str_isascii(input_obj):
            pdf, ld_data = open_pdf_native(input_obj, password)
        else:
            pdf = pdfium.FPDF_LoadDocument(input_obj, password)
            if pdfium.FPDF_GetPageCount(pdf) < 1:
                handle_pdfium_error(False)
        
    elif is_buffer(input_obj):
        pdf, ld_data = open_pdf_buffer(input_obj, password)
        
    else:
        raise ValueError(
            "Input must be a file path, bytes or a byte buffer, but it is {}.".format( type(input_obj) )
        )
    
    return pdf, ld_data
    


def close_pdf(
        pdf: pdfium.FPDF_DOCUMENT,
        loader_data: LoaderData = None,
    ):
    """
    Close an in-memory PDFium document.
    
    Parameters:
        pdf:
            The PDFium document object to close using ``FPDF_CloseDocument()``.
        loader_data:
            A :class:`.LoaderData` object, as returned by :func:`.open_pdf_auto`,
            :func:`.open_pdf_buffer` or :func:`.open_pdf_native`.
    """
    pdfium.FPDF_CloseDocument(pdf)
    if loader_data is not None:
        loader_data.close()



def open_pdf(
        input_obj: Union[str, bytes, BinaryIO],
        password: Union[str, bytes] = None,
    ) -> pdfium.FPDF_DOCUMENT:
    """
    This function is deprecated and only included for backward compatibility.
    Please use :class:`.PdfContext` or :func:`.open_pdf_auto` instead.
    """
    
    warnings.warn(
        "open_pdf() is deprecated and scheduled for removal - please use open_pdf_auto() instead.",
        DeprecationWarning
    )
    
    filepath = None
    data = None
    
    if isinstance(input_obj, str):
        filepath = abspath(input_obj)
    elif isinstance(input_obj, bytes):
        data = input_obj
    elif callable(getattr(input_obj, 'read', None)):
        data = input_obj.read()
    else:
        raise ValueError(
            "Input must be a file path, bytes or a byte buffer, but it is {}.".format( type(input_obj) )
        )
    
    if (filepath is not None) and (data is None):
        pdf = pdfium.FPDF_LoadDocument(filepath, password)
    elif (data is not None) and (filepath is None):
        pdf = pdfium.FPDF_LoadMemDocument(data, len(data), password)
    else:
        assert False
    
    page_count = pdfium.FPDF_GetPageCount(pdf)
    if page_count < 1:
        handle_pdfium_error(False)
    
    return pdf



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import subprocess


problematic = ("pathlib", 'f"', "f'")

def main():
    for string in problematic:
        command = [
            "grep",
            "-r",
            string,
            "--include",
            "*.py",
            "--exclude-dir",
            "sourcebuild",
            "--exclude-dir",
            "data",
        ]
        print(" ".join(command))
        subprocess.run(command, cwd=os.getcwd())
        print()


if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from ctypes import c_float, byref
from typing import Sequence, Callable
from pypdfium2 import _pypdfium as pdfium
from pypdfium2._helpers.error_handler import *


def _get_box(
        page: pdfium.FPDF_PAGE,
        box_function: Callable,
        fallback_function: Callable,
    ) -> Sequence[float]:
    
    left, bottom, right, top = c_float(), c_float(), c_float(), c_float()
    
    ret_code = box_function(
        page,
        byref(left), byref(bottom), byref(right), byref(top),
    )
    if not ret_code:
        return fallback_function(page)
    
    return (left.value, bottom.value, right.value, top.value)


def get_mediabox(page: pdfium.FPDF_PAGE) -> Sequence[float]:
    """
    Get the MediaBox of *page* in PDF canvas units (usually 1/72in).
    Falls back to ANSI A if the page does not define a MediaBox.

    Returns:
        A tuple of four float coordinates.
    """
    
    return _get_box(
        page,
        pdfium.FPDFPage_GetMediaBox,
        lambda _page: (0, 0, 612, 792),
    )


def get_cropbox(page: pdfium.FPDF_PAGE) -> Sequence[float]:
    """
    Get the CropBox of *page* in PDF canvas units (usually 1/72in).
    Falls back to MediaBox if the page does not define a CropBox.

    Returns:
        A tuple of four float coordinates.
    """
    
    return _get_box(
        page,
        pdfium.FPDFPage_GetCropBox,
        get_mediabox,
    )



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from pypdfium2 import _namespace as pdfium
from pypdfium2._cli._parser import ArgParser


def parse_args(argv, prog, desc):
    parser = ArgParser(
        prog = prog,
        description = desc,
    )
    parser.add_argument(
        'input',
        help = "PDF document of which to print the outline",
    )
    parser.add_argument(
        '--max-depth',
        type = int,
        default = 15,
        help = "Maximum recursion depth to consider when parsing the table of contents",
    )
    return parser.parse_args(argv)


def main(argv, prog, desc):
    
    args = parse_args(argv, prog, desc)
    
    with pdfium.PdfContext(args.input) as pdf:
        pdfium.print_toc( pdfium.get_toc(pdf, max_depth=args.max_depth) )



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
from os.path import (
    join,
    basename,
)
import shutil
import setuptools
from glob import glob
from typing import Callable
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel
from _packaging import (
    Libnames,
    SourceTree,
    DataTree,
    ModuleDir,
    extract_version,
)

class PlatformDirs:
    Darwin64     = join(DataTree,'darwin-x64')
    DarwinArm64  = join(DataTree,'darwin-arm64')
    Linux64      = join(DataTree,'linux-x64')
    LinuxArm64   = join(DataTree,'linux-arm64')
    LinuxArm32   = join(DataTree,'linux-arm32')
    Windows64    = join(DataTree,'windows-x64')
    Windows86    = join(DataTree,'windows-x86')
    WindowsArm64 = join(DataTree,'windows-arm64')
    SourceBuild  = join(DataTree,'sourcebuild')


SetupKws = dict(
    version = extract_version('V_PYPDFIUM2'),
)


class BDistBase (_bdist_wheel):
    def finalize_options(self):
        _bdist_wheel.finalize_options(self)
        self.python_tag = 'py3'
        self.plat_name_supplied = True


def _clean():
    
    build_cache    = join(SourceTree,'build')
    bindings_file  = join(ModuleDir,'_pypdfium.py')
    
    libpaths = []
    for name in Libnames:
        libpaths.append( join(ModuleDir, name) )
    
    files = [bindings_file, *libpaths]
    
    if os.path.exists(build_cache):
        shutil.rmtree(build_cache)
    
    for file in files:
        if os.path.exists(file):
            os.remove(file)


def _copy_bindings(platform_dir):
    
    # non-recursively collect all objects from the platform directory
    for src_path in glob(join(platform_dir,'*')):
        
        # copy platform-specific files into the sources, excluding possible directories
        if os.path.isfile(src_path):
            dest_path = join(ModuleDir, basename(src_path))
            shutil.copy(src_path, dest_path)


def build(lib_setup: Callable, platform_dir):
    _clean()
    _copy_bindings(platform_dir)
    lib_setup()
    _clean()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import math
import ctypes
from PIL import Image
from pypdfium2 import _pypdfium as pdfium
from pypdfium2._helpers import utilities
from pypdfium2._helpers.constants import *


def render_page(
        pdf: pdfium.FPDF_DOCUMENT,
        page_index: int = 0,
        scale: float = 1,
        rotation: int = 0,
        colour: int = 0xFFFFFFFF,
        annotations: bool = True,
        greyscale: bool = False,
        optimise_mode: OptimiseMode = OptimiseMode.none,
    ) -> Image.Image:
    """
    Rasterise a single PDF page using PDFium.
    
    Parameters:
        
        pdf:
            A PDFium document (can be obtained with :class:`.PdfContext` or :func:`.open_pdf_auto`).
        
        page_index:
            Zero-based index of the page to render.
        
        scale:
            
            Define the quality (or size) of the image.
            
            By default, one PDF point (1/72in) is rendered to 1x1 pixel. This factor scales the
            number of pixels that represent one point.
            
            Higher values increase quality, file size and rendering duration, while lower values
            reduce them.
            
            Note that UserUnit is not taken into account, so if you are using PyPDFium2 in
            conjunction with an other PDF library, you may want to check for a possible
            ``/UserUnit`` in the page dictionary and multiply this scale factor with it.
        
        rotation:
            Rotate the page by 90, 180, or 270 degrees. Value 0 means no rotation.
        
        colour:
            
            .. _8888 ARGB: https://en.wikipedia.org/wiki/RGBA_color_model#ARGB32
            
            The background colour to use, given as a hexadecimal integer in `8888 ARGB`_ format.
            Defaults to white (``0xFFFFFFFF``). See also :func:`.colour_as_hex`.
        
        annotations:
            Whether to render page annotations.
        
        greyscale:
            Whether to render in greyscale mode (no colours).
        
        optimise_mode:
            Optimise rendering for LCD displays or for printing.
    
    Returns:
        :class:`PIL.Image.Image`
    """
    
    use_alpha = True
    
    if colour is not None:
        alpha_val = hex(colour)[2:4].upper()
        if alpha_val == 'FF':
            use_alpha = False
    
    page_count = pdfium.FPDF_GetPageCount(pdf)
    if not 0 <= page_index < page_count:
        raise IndexError(
            "Page index {} is out of bounds for document with {} pages.".format(page_index, page_count)
        )
    
    form_config = pdfium.FPDF_FORMFILLINFO(2)
    form_fill = pdfium.FPDFDOC_InitFormFillEnvironment(pdf, form_config)
    
    page = pdfium.FPDF_LoadPage(pdf, page_index)
    pdfium.FORM_OnAfterLoadPage(page, form_fill)
    
    width  = math.ceil(pdfium.FPDF_GetPageWidthF(page)  * scale)
    height = math.ceil(pdfium.FPDF_GetPageHeightF(page) * scale)
    
    if rotation in (90, 270):
        width, height = height, width
    
    bitmap = pdfium.FPDFBitmap_Create(width, height, int(use_alpha))
    if colour is not None:
        pdfium.FPDFBitmap_FillRect(bitmap, 0, 0, width, height, colour)
    
    render_flags = 0x00
    
    if annotations:
        render_flags |= pdfium.FPDF_ANNOT
    if greyscale:
        render_flags |= pdfium.FPDF_GRAYSCALE
    
    if optimise_mode is OptimiseMode.none:
        pass
    elif optimise_mode is OptimiseMode.lcd_display:
        render_flags |= pdfium.FPDF_LCD_TEXT
    elif optimise_mode is OptimiseMode.printing:
        render_flags |= pdfium.FPDF_PRINTING
    else:
        raise ValueError("Invalid optimise_mode {}".format(optimise_mode))
    
    render_args = [
        bitmap,
        page,
        0, 0,
        width, height,
        utilities.translate_rotation(rotation),
        render_flags,
    ]
    
    pdfium.FPDF_RenderPageBitmap(*render_args)
    pdfium.FPDF_FFLDraw(form_fill, *render_args)
    
    cbuffer = pdfium.FPDFBitmap_GetBuffer(bitmap)
    buffer = ctypes.cast(cbuffer, ctypes.POINTER(ctypes.c_ubyte * (width * height * 4)))
    
    pil_image = Image.frombuffer("RGBA", (width, height), buffer.contents, "raw", "BGRA", 0, 1)
    
    if greyscale:
        if use_alpha:
            pil_image = pil_image.convert("LA")
        else:
            pil_image = pil_image.convert("L")
    
    elif not use_alpha:
        pil_image = pil_image.convert("RGB")
    
    pdfium.FPDFBitmap_Destroy(bitmap)
    pdfium.FORM_OnBeforeClosePage(page, form_fill)
    pdfium.FPDF_ClosePage(page)
    pdfium.FPDFDOC_ExitFormFillEnvironment(form_fill)
    
    return pil_image



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import sysconfig
from os.path import (
    join,
    basename,
)

import update_pdfium
import build_pdfium
import setup_source
import setup_darwin_arm64
import setup_darwin_x64
import setup_linux_arm32
import setup_linux_arm64
import setup_linux_x64
import setup_windows_arm64
import setup_windows_x64
import setup_windows_x86
import _getdeps as getdeps
from _packaging import DataTree
from _setup_base import PlatformDirs


StatusFile = join(DataTree, 'setup_status.txt')

def check_presetup() -> bool:
    with open(StatusFile, 'r') as file_handle:
        content = file_handle.read().strip()
    if content == 'InitialState':
        return True
    elif content == 'PreSetupDone':
        return False
    else:
        raise ValueError( "Invalid content in setup status file: '{}'".format(content) )


class PlatformManager:
    
    def __init__(self):
        
        plat_name = sysconfig.get_platform().lower()
        for char in ('-', '.'):
            plat_name = plat_name.replace(char, '_')
        
        self.plat_name = plat_name
    
    def _is_platform(self, start, end):
        if self.plat_name.startswith(start):
            if self.plat_name.endswith(end):
                return True
        return False
    
    def is_darwin_arm64(self):
        return self._is_platform('macosx', 'arm64')
    
    def is_darwin_x64(self):
        return self._is_platform('macosx', 'x86_64')
    
    def is_linux_arm32(self):
        return self._is_platform('linux', 'armv7l')
    
    def is_linux_arm64(self):
        return self._is_platform('linux', 'aarch64')
    
    def is_linux_x64(self):
        return self._is_platform('linux', 'x86_64')
    
    def is_windows_arm64(self):
        return self._is_platform('win', 'arm64')
    
    def is_windows_x64(self):
        return self._is_platform('win', 'amd64')
    
    def is_windows_x86(self):
        return self._is_platform('win32', '')


# function to generate bindings, if doing pre-setup
def _make_bindings(platform_dir, w_presetup):
    if w_presetup:
        update_pdfium.main( ['-p', basename(platform_dir)] )


def main():
    
    # Since setuptools may run this code multiple times with different commands,
    # we have a status file to check whether pre-setup tasks have already been done.
    # If you deliberately wish to re-run them, set the content of `data/setup_status.txt`
    # to `InitialState`.
    w_presetup = check_presetup()
    
    if w_presetup:
        
        # automatically check/install dependencies
        getdeps.main()
        
        # update status file
        with open(StatusFile, 'w') as file_handle:
            file_handle.write("PreSetupDone")
    
    # tooling to determine the current platform
    plat = PlatformManager()
    
    # run the corresponding setup code
    if plat.is_darwin_arm64():
        _make_bindings(PlatformDirs.DarwinArm64, w_presetup)
        setup_darwin_arm64.main()
    
    elif plat.is_darwin_x64():
        _make_bindings(PlatformDirs.Darwin64, w_presetup)
        setup_darwin_x64.main()
    
    elif plat.is_linux_arm32():
        _make_bindings(PlatformDirs.LinuxArm32, w_presetup)
        setup_linux_arm32.main()
    
    elif plat.is_linux_arm64():
        _make_bindings(PlatformDirs.LinuxArm64, w_presetup)
        setup_linux_arm64.main()
    
    elif plat.is_linux_x64():
        _make_bindings(PlatformDirs.Linux64, w_presetup)
        setup_linux_x64.main()
    
    elif plat.is_windows_arm64():
        _make_bindings(PlatformDirs.WindowsArm64, w_presetup)
        setup_windows_arm64.main()
    
    elif plat.is_windows_x64():
        _make_bindings(PlatformDirs.Windows64, w_presetup)
        setup_windows_x64.main()
    
    elif plat.is_windows_x86():
        _make_bindings(PlatformDirs.Windows86, w_presetup)
        setup_windows_x86.main()
    
    # Platform without pre-built binaries - trying a regular sourcebuild
    # In case it does not work, you may want to attempt a native build (`./build_pdfium -p`),
    # and then write a wheel to `dist/` using `python3 setup_source.py bdist_wheel`
    else:
        if w_presetup:
            args = build_pdfium.parse_args([])
            build_pdfium.main(args)
        setup_source.main()


if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause 

import subprocess
from os.path import (
    expanduser,
    dirname,
    abspath,
    join,
)

Libnames = [
    'libpdfium.dylib',
    'pdfium.dylib',
    'libpdfium.dll',
    'pdfium.dll',
    'libpdfium.so',
    'pdfium.so',
    'libpdfium',
    'pdfium',
]


HomeDir     = expanduser('~')
SourceTree  = dirname(abspath(__file__))
DataTree    = join(SourceTree,'data')
SB_Dir      = join(SourceTree,'sourcebuild')
ModuleDir   = join(SourceTree,'src','pypdfium2')
VersionFile = join(ModuleDir,'_version.py')


def run_cmd(command, cwd):
    print(command)
    subprocess.run(command, cwd=cwd, shell=True)


def postprocess_bindings(bindings_file, platform_dir):
    
    with open(bindings_file, 'r') as file_reader:
        text = file_reader.read()
        #text = text.split('\n"""\n', maxsplit=1)[1]
        text = text.replace(platform_dir, '.')
        text = text.replace(HomeDir, '~')
    
    with open(bindings_file, 'w') as file_writer:
        file_writer.write(text)


def _get_ver_namespace():
    
    ver_namespace = {}
    with open(VersionFile) as fh:
        exec(fh.read(), ver_namespace)
    
    return ver_namespace


_ver_namespace = _get_ver_namespace()

def extract_version(variable_str):
    return _ver_namespace[variable_str]
